package com.serval.polsplitter;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.Map;

/**
 * User: tej
 * Date: Apr 4, 2011
 * Time: 3:36:46 PM
 */
public class PolicyAnalysis {


     @SuppressWarnings({"UnusedDeclaration"})
    public static void buildStats(String[] args) throws FileNotFoundException {

        StringBuffer htmlCode = new StringBuffer("<html><body>");

        for (String arg : args) {

            htmlCode.append(readRulesFromXACML(arg));

            System.out.println("____________________________________________________________________________________________");

        }

        htmlCode.append("</html></body>");

        // write html code to file

        PrintWriter pw = new PrintWriter("stats.html");
        pw.println(htmlCode.toString());
        pw.flush();
        pw.close();

    }


        /**
     * Reads rules from the xacml policies, reads only (rule_type,role,action,resource)
     *
     * @param filename the name of the file to analyze
     * @return the html code with stats
     */
    public static String readRulesFromXACML(String filename) {

        String subject, action, resource, action_resource;

        Map<String, Integer> rolesCount, actionResourceCount;
        rolesCount = new HashMap<String, Integer>();
        actionResourceCount = new HashMap<String, Integer>();

        String htmlCode = "";


        try {
            File file = new File(filename);

            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            Document doc = db.parse(file);
            doc.getDocumentElement().normalize();

            NodeList nodeLst = doc.getElementsByTagName("Rule");
            System.out.println("Policy = " + file.getName());

            // if no rule, check if there are at least a policy or policySet in which there is one rule
            if (nodeLst.getLength() == 0) {

                nodeLst = doc.getElementsByTagName("PolicySet");
            }


            // print out nb rules
            System.out.println("NB Rules = " + nodeLst.getLength());

            htmlCode = "<h1> Policy : " + file.getName() + "   ,  # rules  = " + nodeLst.getLength() + "</h1>";


            for (int s = 0; s < nodeLst.getLength(); s++) {

                Node fstNode = nodeLst.item(s);

                // subject
                Element ruleElmnt = (Element) fstNode;
                subject = extractAttribValue(ruleElmnt.getElementsByTagName("Subject"), "Subject");
                action = extractAttribValue(ruleElmnt.getElementsByTagName("Action"), "Action");
                resource = extractAttribValue(ruleElmnt.getElementsByTagName("Resource"), "Resource");

                //System.out.println("Rule(" + subject + " , " + action + " , " + resource + ")");

                //System.out.println("******************************************************************************");

                // ************************* Do analysis on the fly


                // increment rolecount if already there, create if not
                if (rolesCount.containsKey(subject)) {
                    rolesCount.put(subject, rolesCount.get(subject) + 1);
                } else {
                    rolesCount.put(subject, 1);
                }


                action_resource = action + "_" + resource;

                // increment actionResourceCount if already there, create if not
                if (actionResourceCount.containsKey(action_resource)) {
                    actionResourceCount.put(action_resource, actionResourceCount.get(action_resource) + 1);

                } else {
                    actionResourceCount.put(action_resource, 1);

                }

                // ************************* end analysis on the fly

            }


            // ******************************************** print results
            System.out.println("Roles  *************************************************************************************");


            // print out count of roles and then of action_resource
            htmlCode += "<br/><table border=5><tr><td><b>Role</b></td><td><b># rules</b></td></tr>";

            for (String role : rolesCount.keySet()) {
                System.out.println(role + " ===> " + rolesCount.get(role));

                htmlCode += "<tr><td><b>" + role + "</b></td><td>" + rolesCount.get(role) + "</td></tr>";


            }
            htmlCode += "</table>";

            System.out.println("Action_resource  ***************************************************************************");

            htmlCode += "<br/><br/><table border=5><tr><td><b>Action_Resource</b></td><td><b># rules</b></td></tr>";

            for (String a_r : actionResourceCount.keySet()) {
                System.out.println(a_r + " ===> " + actionResourceCount.get(a_r));

                htmlCode += "<tr><td><b>" + a_r + "</b></td><td>" + actionResourceCount.get(a_r) + "</td></tr>";

            }

            htmlCode += "</table>";


        } catch (ParserConfigurationException e) {
            e.printStackTrace();
        } catch (SAXException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }

        return htmlCode;

    }


    /**
        * extract the attributes
        *
        * @param nl      the upper nodeList of the rule
        * @param tagName the tagname subject, action or resource
        * @return the attribute value of the subject, the action or the resource
        */

       private static String extractAttribValue(NodeList nl, String tagName) {

           // return AnySbject/AnyResource/AnyAction if item nb = 0

           if (nl.getLength() == 0)
               return "Any" + tagName;

           // AttributeValue
           Node attrNode = nl.item(0);

           Element attElmnt = (Element) attrNode;

           NodeList attvalue = attElmnt.getElementsByTagName("AttributeValue");

           return ((NodeList) attvalue.item(0)).item(0).getNodeValue();

       }


    /**
     * Provide analysis on how the policy will be distributed
     *
     * @param rules the rules of the policy
     */
    public static void policyAnalysis(String[][] rules) {


        Map<String, Integer> rolesCount, actionResourceCount;

        rolesCount = new HashMap<String, Integer>();
        actionResourceCount = new HashMap<String, Integer>();
        String action_resource;


        // calculate number of roles and fill in the roleCount map & the actionResourceCount map
        // role is int index 2 & action_resource are in 3 & 4

        for (String[] rule : rules) {

            // increment rolecount if already there, create if not
            if (rolesCount.containsKey(rule[2])) {
                rolesCount.put(rule[2], rolesCount.get(rule[2]) + 1);
            } else {
                rolesCount.put(rule[2], 1);
            }


            action_resource = rule[3] + "_" + rule[4];

            // increment actionResourceCount if already there, create if not
            if (actionResourceCount.containsKey(action_resource)) {
                actionResourceCount.put(action_resource, actionResourceCount.get(action_resource) + 1);

            } else {
                actionResourceCount.put(action_resource, 1);
            }
        }


        System.out.println("Roles  *************************************************************************************");


        // print out count of roles and then of action_resource

        for (String role : rolesCount.keySet()) {
            System.out.println(role + " ===> " + rolesCount.get(role));
        }


        System.out.println("Action_resource  ***************************************************************************");


        for (String a_r : actionResourceCount.keySet()) {
            System.out.println(a_r + " ===> " + actionResourceCount.get(a_r));
        }
    }
}
