package com.serval.polsplitter;

import com.serval.polsplitter.structs.Target;
import com.serval.polsplitter.structs.TargetElement;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import javax.xml.XMLConstants;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;
import java.io.*;
import java.util.*;
import java.util.regex.Pattern;

/**
 * User: tej
 * Date: 4/6/11
 * Time: 10:15 AM
 */
public class Utils {


    public static boolean searching = true;
    public static Collection<String> listSearchResults = new ArrayList<String>();

    /**
     * Calculates all Clustering Criteria (CC)
     *
     * @param targets targets to be used for clustering
     * @return all clustering criteria lists in a map
     */
    public static Map<Integer, Collection<Target>> computeAllClusteringCriteria(Collection<Target> targets) {

        Map<Integer, Collection<Target>> result;
        Collection<Target> ccsSubjects, ccsActions, ccsResources, ccsSubjectAction, ccsSubjectResource, ccsActionsResource, ccsSubjectActionResource;
        Target targetToAdd;
        TargetElement targetElement;

        Collection<TargetElement> elements;


        result = new HashMap<Integer, Collection<Target>>();

        ccsSubjects = new HashSet<Target>();
        ccsActions = new HashSet<Target>();
        ccsResources = new HashSet<Target>();

        // all two
        ccsSubjectAction = new HashSet<Target>();
        ccsSubjectResource = new HashSet<Target>();
        ccsActionsResource = new HashSet<Target>();

        // all three
        ccsSubjectActionResource = new HashSet<Target>();

        // start by getting all subjects, actions & resources collections
        for (Target target : targets) {

            // add all subjects, actions & resources
            elements = target.getSubjects();

            for (TargetElement element : elements) {
                targetElement = element;

                // build target with that CC
                targetToAdd = new Target();
                targetToAdd.getSubjects().add(targetElement);

                ccsSubjects.add(targetToAdd);

            }

            elements = target.getActions();

            for (TargetElement element : elements) {
                targetElement = element;

                // build target with that CC
                targetToAdd = new Target();
                targetToAdd.getActions().add(targetElement);

                ccsActions.add(targetToAdd);
            }

            elements = target.getResources();

            for (TargetElement element : elements) {
                targetElement = element;

                // build target with that CC
                targetToAdd = new Target();
                targetToAdd.getResources().add(targetElement);

                ccsResources.add(targetToAdd);
            }
        }

        result.put(PolicySplitter.SUBJECT, ccsSubjects);
        result.put(PolicySplitter.ACTION, ccsActions);
        result.put(PolicySplitter.RESOURCE, ccsResources);

        // new get all combinations of 2
        for (Target target : targets) {

            ccsSubjectAction.addAll(allCombinationOfTwoTargetElements(target, PolicySplitter.SUBJECT_ACTION));
            ccsSubjectResource.addAll(allCombinationOfTwoTargetElements(target, PolicySplitter.SUBJECT_RESOURCE));
            ccsActionsResource.addAll(allCombinationOfTwoTargetElements(target, PolicySplitter.ACTION_RESOURCE));

        }

        // now for all three
        for (Target target : targets) {

            ccsSubjectActionResource.addAll(allCombinationOfThreeTargetElements(target));
        }

        // put all in result
        result.put(PolicySplitter.SUBJECT, ccsSubjects);
        result.put(PolicySplitter.ACTION, ccsActions);
        result.put(PolicySplitter.RESOURCE, ccsResources);

        result.put(PolicySplitter.SUBJECT_ACTION, ccsSubjectAction);
        result.put(PolicySplitter.SUBJECT_RESOURCE, ccsSubjectResource);
        result.put(PolicySplitter.ACTION_RESOURCE, ccsActionsResource);

        result.put(PolicySplitter.SUBJECT_ACTION_RESOURCE, ccsSubjectActionResource);


        return result;
    }

    /**
     * Computes and 2 possible combination of 2 parameters in the given target
     *
     * @param target the target with the elements to combine
     * @param cc     the clustering criteria used to combine
     * @return list on targets with only one conbination each
     */
    static Collection<Target> allCombinationOfTwoTargetElements(Target target, int cc) {

        Collection<Target> result;
        Target targetTmp;
        Collection<TargetElement> firstElements, secondEelements;

        result = new ArrayList<Target>();


        switch (cc) {

            case PolicySplitter.SUBJECT_ACTION:
                firstElements = target.getSubjects();
                secondEelements = target.getActions();
                break;
            case PolicySplitter.SUBJECT_RESOURCE:
                firstElements = target.getSubjects();
                secondEelements = target.getResources();
                break;
            case PolicySplitter.ACTION_RESOURCE:
                firstElements = target.getActions();
                secondEelements = target.getResources();
                break;
            default:
                return null;

        }

        // mix
        for (TargetElement targetElement1 : firstElements) {
            for (TargetElement targetElement2 : secondEelements) {

                targetTmp = new Target();

                switch (cc) {

                    case PolicySplitter.SUBJECT_ACTION:
                        targetTmp.getSubjects().add(targetElement1);
                        targetTmp.getActions().add(targetElement2);
                        break;
                    case PolicySplitter.SUBJECT_RESOURCE:
                        targetTmp.getSubjects().add(targetElement1);
                        targetTmp.getResources().add(targetElement2);
                        break;
                    case PolicySplitter.ACTION_RESOURCE:
                        targetTmp.getActions().add(targetElement1);
                        targetTmp.getResources().add(targetElement2);
                        break;

                }

                result.add(targetTmp);

            }


        }

        return result;
    }

    /**
     * Compute oall combination of (Subject,Action,Resource)
     *
     * @param target the target to use
     * @return the list of all combination in separate targets
     */
    static Collection<Target> allCombinationOfThreeTargetElements(Target target) {

        Collection<Target> result;
        Target targetTmp;

        Collection<TargetElement> subjects, actions, resources;


        result = new ArrayList<Target>();

        subjects = target.getSubjects();
        actions = target.getActions();
        resources = target.getResources();

        for (TargetElement subject : subjects) {
            for (TargetElement action : actions) {
                for (TargetElement resource : resources) {

                    targetTmp = new Target();
                    targetTmp.getSubjects().add(subject);
                    targetTmp.getActions().add(action);
                    targetTmp.getResources().add(resource);

                    result.add(targetTmp);

                }

            }

        }


        return result;
    }

    /**
     *
      * @param rootDir the root search dir
     * @param fileList list of xacml files
     */
 private static void searchForXMLFiles(String rootDir,String[] fileList)
 {

  for (int i = 0; i < fileList.length && searching; i++)
  {
   File f = new File(rootDir,fileList[i]);

   if (f.isDirectory())
   {
    String path = f.getPath();
    searchForXMLFiles(path, f.list());
   }
   else
   {


    if(f.isFile() && f.getName().contains(".")) {

    String extension = f.getName().substring(f.getName().lastIndexOf(".") + 1 , f.getName().length());

    if (extension.equalsIgnoreCase("xml"))
        listSearchResults.add(f.getAbsolutePath());
   }
   }
  }
 }

    public static boolean checkXMLisXACML(String filename) {

        try {

        SchemaFactory factory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
        Schema schema = factory.newSchema(new File("/home/tej/IdeaProjects/xacmlTool/cs-xacml-schema-policy-01.xsd"));
        Validator validator = schema.newValidator();

        DocumentBuilder parser = DocumentBuilderFactory.newInstance().newDocumentBuilder();
        Document document = parser.parse(new File(filename));

        validator.validate(new DOMSource(document));


        } catch (SAXException e) {
            System.out.println(filename);
            System.out.println(filename + " is not a valid XACML 1.0");

           e.printStackTrace();

            // not valid
            return false;

        } catch (IOException e) {
           // System.out.println(filename);
           //  e.printStackTrace();

            // problem in handling file - return false too
            return false;
        } catch (ParserConfigurationException e) {
           // System.out.println(filename);
           // e.printStackTrace();

           // problem in handling file - return false too
            return false;
        }


        System.out.println(filename + " is a valid XACML 1.0");
        return true;
    }

    public static boolean checkXMLisXACMLWithXMLStarlet(String filename) {

        Process p;
        boolean isValid = false;


        try {
            p = Runtime.getRuntime().exec("xmlstarlet val -s /home/tej/IdeaProjects/xacmlTool/cs-xacml-schema-policy-01.xsd" +
                    " " + filename);

            BufferedReader br = new BufferedReader(new InputStreamReader(p.getInputStream()));
            String isValidStr = br.readLine();

            isValid = !isValidStr.contains("invalid");


            br.close();


        } catch (IOException e) {
            e.printStackTrace();
        }


        return isValid;

    }



    public static Collection<String> findXACMLsInDirectory(String directory) {


        listSearchResults = new ArrayList<String>();
        searchForXMLFiles(directory, new File(directory).list());


        return listSearchResults;
    }


    public static void main(String[] args) {

        File ff = new File("/home/tej/tmp/xacml-subjects_another");
        searchForXMLFiles(ff.getPath(),ff.list());

        System.out.println("Searching ...");
        System.out.println("Search result : ");

        int nb = 0;
        for (String xacmlFile : listSearchResults) {

        boolean isValid = checkXMLisXACMLWithXMLStarlet(xacmlFile);

            if(!isValid) {
                System.out.println(xacmlFile);
                nb++;
            }
        }

        System.out.println("NB = " + nb);

    }


    /**
     * transforms the XACML File by putting in the policy or policySet Target Subject or Action or Resource from CC
     * @param filename XACML file to be processed
     * @param targetElement the targetElement type (S,A or R)
     * @param cluster the CC element value
     */
    private static void transformXACMLFile(String filename, String targetElement, String cluster) {

        // check parameters
        if (cluster == null || cluster.equals(""))
            return;

        //System.out.println(filename);

        CharSequence subject = "S";
        CharSequence action = "A";
        CharSequence resource = "R";
        String chaine_rep = "";
        String index_debut = "";
        String index_fin = "";
        String tagName = "";
        int insertion_debut = 0;
        int index_supp = 0;

        if (targetElement.equals(subject)) {

            //System.out.println(cluster);
            tagName = "Subjects";
            index_debut = "<Subjects>";
            index_fin = "</Subjects>";
            chaine_rep = "<AnySubject/>";
            insertion_debut = 10;
            index_supp = 13;

        } else if (targetElement.equals(action)) {
            tagName = "Actions";
            index_debut = "<Actions>";
            index_fin = "</Actions>";
            chaine_rep = "<AnyAction/>";
            insertion_debut = 9;
            index_supp = 12;
        } else if (targetElement.equals(resource)) {
            tagName = "Resources";
            index_debut = "<Resources>";
            index_fin = "</Resources>";
            chaine_rep = "<AnyResource/>";
            insertion_debut = 11;
            index_supp = 14;
        }


        try {

            File file = new File(filename);
            String monFile = "";
            String MonFile1;
            String res = "";
            boolean result = false;

            //  boolean test =true;
            if (file.exists()) {

                DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
                DocumentBuilder builder = factory.newDocumentBuilder();

                Document doc = builder.parse(filename);
                TransformerFactory tranFactory = TransformerFactory.newInstance();
                Transformer aTransformer = tranFactory.newTransformer();

                // Get nodes list of all elements
                NodeList list = doc.getElementsByTagName(tagName);
                for (int i = 0; i < list.getLength(); i++) {
                    if (!result) { // Get element
                        Element element = (Element) list.item(i);
                        Source src = new DOMSource(element);

                        // System.out.println("Node no: " + i + " is " + element.getNodeName());

                        Result dest = new StreamResult("temp.xml");
                        aTransformer.transform(src, dest);
                        monFile = readFileAsString("temp.xml");
                        // monFile = PolicySplitter.removeReservedCharsFromString(monFile);

                        // System.out.println(monFile);

                        // remove illegal char and see if equals
                        result = monFile.contains(cluster);
                        // System.out.println(result);
                    }
                    // System.out.println(monFile);
                    int debut = monFile.indexOf(index_debut);
                    int fin = monFile.indexOf(index_fin);
                    res = monFile.substring(debut + insertion_debut, fin);
                    // System.out.println(res);
                }

                MonFile1 = readFileAsString(filename);
                int debut1 = MonFile1.indexOf(index_debut);
                int deletion = MonFile1.indexOf(chaine_rep);
                // System.out.println(debut1);
                StringBuffer MonFile_b = new StringBuffer(MonFile1);
                StringBuffer res_b = new StringBuffer(res);
                MonFile_b.delete(deletion, deletion + index_supp);
                MonFile_b.insert(debut1 + insertion_debut, res_b);
                // MonFile_b.insert(debut1 + insertion_debut, res_b);
                // System.out.println(MonFile_b);

                if (result) {

                    EcrireFichierXML(MonFile_b.toString(), filename);
                }
            } else {
                // System.out.println(filename + " (file name) doesn't found!");
            }
        } catch (Exception e) {
            e.getMessage();
        }
    }

    /**
     * reads and puts file in a string
     * @param filePath  file to read
     * @return string of the file content
     * @throws java.io.IOException when occurs
     */
    private static String readFileAsString(String filePath)
            throws java.io.IOException {
        StringBuffer fileData = new StringBuffer(1000);
        BufferedReader reader = new BufferedReader(
                new FileReader(filePath));
        char[] buf = new char[1024];
        int numRead;
        while ((numRead = reader.read(buf)) != -1) {
            String readData = String.valueOf(buf, 0, numRead);
            fileData.append(readData);
            buf = new char[1024];
        }
        reader.close();
        return fileData.toString();
    }

    /**
     * writes a string inside a file
     * @param string the content to write
     * @param final_output file to write inside
     * @throws IOException when occurs, who knows
     */
    private static void EcrireFichierXML(String string, String final_output) throws IOException {

        PrintWriter ecrivain;
        ecrivain = new PrintWriter(new BufferedWriter
                (new FileWriter(final_output)));
        ecrivain.println(string);
        ecrivain.close();
    }


    /**
     * extracts the CC tags from fileName
     * @param fileName the XACML file name
     * @return array with the 3 values extracted, value is "" if not there
     */
    private static String[] extractTags(String fileName) {

        String subject = "", action = "", resource = "";

        String tag = "", tags[] = {"_S_", "_A_", "_R_", "_SA_", "_SR_", "_AR_", "_SAR_"};
        String[][] elements = {{subject}, {action}, {resource}, {subject, action}, {subject, resource}, {action, resource}, {subject, action, resource}};

        Pattern pattern = Pattern.compile("_");
        String[] items = pattern.split(fileName.substring(0, fileName.lastIndexOf(".")));
        int tagIndex = 0;


        for (int i = 0; i < tags.length; i++) {
            if (fileName.contains(tags[i])) {
                tag = tags[i];
                tagIndex = i;
                break;

            }
        }

        int i = 1;
        while (i <= tag.length() - 2) {

            // System.out.println(items[items.length - i]);

            elements[tagIndex][elements[tagIndex].length - i] = new String(items[items.length - i]);
            i++;
        }

        // show the three items
        // System.out.println("Subject ===> " + subject);
        //System.out.println("Action ===> " + action);
        //System.out.println("Resource ===> " + resource);

        if (tag.equals("_S_")) {

            subject = elements[tagIndex][0];

        } else if (tag.equals("_A_")) {

            action = elements[tagIndex][0];

        } else if (tag.equals("_R_")) {

            resource = elements[tagIndex][0];

        } else if (tag.equals("_SA_")) {

            subject = elements[tagIndex][0];
            action = elements[tagIndex][1];

        } else if (tag.equals("_AR_")) {

            action = elements[tagIndex][0];
            resource = elements[tagIndex][1];

        } else if (tag.equals("_SR_")) {

            subject = elements[tagIndex][0];
            resource = elements[tagIndex][1];

        } else if (tag.equals("_SAR_")) {

            subject = elements[tagIndex][0];
            action = elements[tagIndex][1];
            resource = elements[tagIndex][2];
        }


        return new String[]{subject, action, resource};
    }

    /**
     * transforms the XACML File by putting in the policy or policySet Target elements from CC
     * @param fileName the file to process
     */
    public static void updateTargetInPolicyFile(String fileName) {


        String tags[] = extractTags(fileName);

        // do transformation
        transformXACMLFile(fileName, "S", tags[0]);
        transformXACMLFile(fileName, "A", tags[1]);
        transformXACMLFile(fileName, "R", tags[2]);

    }
}
